@@ -2,24 +2,30 @@ import functools
 import numbers
 from bisect import bisect_left
 from collections import UserDict, namedtuple
-from collections.abc import MutableMapping
+from collections.abc import MutableMapping, Mapping
 from copy import deepcopy
 from operator import attrgetter
-from typing import Any, Callable, Dict, Hashable, Iterator, List, Optional, Tuple, TypeVar, Union, Sequence
+from typing import Any, Callable, Dict, Hashable, Iterator, List, Optional, Tuple, TypeVar, Union, Sequence, Set
 
 T = TypeVar("T")
 K = TypeVar("K")
 V = TypeVar("V")
+HashableT = TypeVar("HashableT", bound=Hashable)
 
 
-def dict_until(obj, keys: list, terminate: Optional[Callable[[T], bool]] = None, default=None) -> T:
+def dict_until(obj, keys: List[HashableT],
+               terminate: Optional[Callable[[Any], bool]] = None,
+               default: Optional[Any] = None) -> Any:
     class Empty:
         pass
 
     UNSIGNED = Empty()
 
+    def default_terminate(v: Any) -> bool:
+        return v is not UNSIGNED
+
     if terminate is None:
-        def terminate(v): return v is not UNSIGNED
+        terminate = default_terminate
 
     from pythonic_toolbox.utils.list_utils import until
 
@@ -31,10 +37,10 @@ def dict_until(obj, keys: list, terminate: Optional[Callable[[T], bool]] = None,
     return val
 
 
-def collect_leaves(data: Optional[Union[dict, List]] = None,
-                   keypath_pred: Optional[Callable[[List[Hashable]], bool]] = None,
-                   leaf_pred: Optional[Callable[[T], bool]] = None) -> List[T]:
-    leaves = list()
+def collect_leaves(data: Optional[Union[Dict, List]] = None,
+                   keypath_pred: Optional[Callable[[List[HashableT]], bool]] = None,
+                   leaf_pred: Optional[Callable[[Any], bool]] = None) -> List[Any]:
+    leaves: List[Any] = list()
     if not data:
         return leaves
 
@@ -61,19 +67,21 @@ def collect_leaves(data: Optional[Union[dict, List]] = None,
     return leaves
 
 
-def select_list_of_dicts(dict_lst: List[dict],
-                         preds: Optional[Iterator[Callable[[dict], bool]]] = None,
-                         keys: Optional[Hashable] = None,
-                         unique=False, val_for_missing_key=None) -> List[dict]:
+def select_list_of_dicts(dict_lst: List[Dict],
+                         preds: Optional[List[Callable[[Dict], bool]]] = None,
+                         keys: Optional[List[HashableT]] = None,
+                         unique=False, val_for_missing_key=None) -> List[Dict]:
     """ Select part of the dict collections."""
 
     from funcy import rpartial, project, all_fn
 
     preds = preds or []
+    preds = list(preds)  # make a shallow copy
     keys = keys or []
+    keys = list(keys)  # make a shallow copy
 
     dict_lst = deepcopy(dict_lst)
-    res: Union[List[dict], Iterator[dict]] = dict_lst
+    res: Union[List[Dict], Iterator[Dict]] = dict_lst
 
     if preds:
         res = filter(all_fn(*preds), dict_lst)
@@ -85,13 +93,13 @@ def select_list_of_dicts(dict_lst: List[dict],
         res = [{k: dct.get(k, val_for_missing_key) for k in keys} for dct in res]
 
     if unique is True:
-        res = unique_list_of_dicts(res)
+        res = list(unique_list_of_dicts(res))
 
     return list(res)
 
 
-def unique_list_of_dicts(dict_list: List[dict]) -> List[dict]:
-    unique_res: List[dict] = []
+def unique_list_of_dicts(dict_list: List[Dict]) -> List[Dict]:
+    unique_res: List[Dict] = list()
     items_tuple_set = set()
     for d in dict_list:
         items_tuple = tuple(d.items())
@@ -101,9 +109,9 @@ def unique_list_of_dicts(dict_list: List[dict]) -> List[dict]:
     return unique_res
 
 
-def walk_leaves(data: Optional[Union[dict, List]] = None,
+def walk_leaves(data: Optional[Union[Dict, List]] = None,
                 trans_fun: Optional[Callable[[Any], Any]] = None,
-                inplace=False) -> Optional[Union[dict, List]]:
+                inplace: bool = False) -> Optional[Union[Dict, List]]:
     """
     :param data: data can be nested dict, list
     :param trans_fun: leaf transform function
@@ -217,7 +225,7 @@ class MyUserDict(MutableMapping):
 
 
 class DictObj(MyUserDict):
-    def __init__(self, in_dict: dict):
+    def __init__(self, in_dict: Dict):
 
         in_dict = deepcopy(in_dict)
 
@@ -287,6 +295,15 @@ class DictObj(MyUserDict):
         except KeyError:
             raise AttributeError
 
+    def __eq__(self, other: 'DictObj') -> bool:
+        if isinstance(other, DictObj):
+            return self._user_dict_hidden_data == other._user_dict_hidden_data
+        return False
+
+    def __hash__(self):
+        """not hashable"""
+        return None
+
     def to_dict(self):
         result = {}
         for key, item in self._user_dict_hidden_data.items():
@@ -314,7 +331,7 @@ class FinalDictObj(DictObj):
     __is_frozen = False
     __frozen_err_msg = 'Cannot modify attribute/item in an already initialized FinalDictObj'
 
-    def __init__(self, in_dict: dict):
+    def __init__(self, in_dict: Dict):
 
         in_dict = deepcopy(in_dict)
 
@@ -389,6 +406,13 @@ class RangeKeyDict:
         def __repr__(self):
             return f'RangeKeyDict.Segment(begin={repr(self.begin)}, end={repr(self.end)}, val={repr(self.val)}'
 
+        def __eq__(self, other: 'RangeKeyDict.Segment') -> bool:
+            return self.begin == other.begin and self.end == other.end and self.val == other.val
+
+        def __hash__(self):
+            """not hashable"""
+            return None
+
     def __init__(self, input_dict: Dict[Union[Tuple[K, K], K], V]) -> None:
         """keys for input dict must be tuple-like intervals (left-closed, right-open) or single point"""
         # input validation and generate inner-used structures
@@ -447,7 +471,7 @@ class RangeKeyDict:
                     raise ValueError(f'Invalid key for {repr(key)}, '
                                      f'left boundary keys must <= right boundary key, '
                                      f'and both of them must be hashable, have completed comparison methods')
-            elif isinstance(key, Hashable):
+            elif not isinstance(key, tuple) and isinstance(key, Hashable):
                 single_point_map[key] = val
                 boundary_keys.append(key)
                 left_boundary_key = right_boundary_key = key
@@ -469,6 +493,13 @@ class RangeKeyDict:
 
         return single_point_map, left_boundary_key_segment_map, segments
 
+    def __eq__(self, other: object) -> bool:
+        if not isinstance(other, RangeKeyDict):
+            return False
+        return (self._single_point_map == other._single_point_map and
+                self._left_boundary_segment_map == other._left_boundary_segment_map and
+                self._sorted_segments == other._sorted_segments)
+
     def __getitem__(self, number):
         if number in self._single_point_map:
             return self._single_point_map[number]
@@ -521,7 +552,7 @@ class StrKeyIdDict(UserDict):
         raw_data.update(my_dict)
         raw_data.update(kwargs)
 
-        valid_data = {}
+        valid_data: Dict[str, Any] = {}
         for key, val in raw_data.items():
             if not self.is_valid_key(key):
                 raise TypeError(
@@ -530,7 +561,7 @@ class StrKeyIdDict(UserDict):
                 valid_data[str(key)] = val
             else:
                 # handle duplicated keys (e.g. '1', 1)
-                duplicate_keys = set()
+                duplicate_keys: Set[Union[str, int]] = set()
                 if str(key) in raw_data.keys():
                     duplicate_keys.add(str(key))
                 if int(key) in raw_data.keys():
@@ -559,6 +590,20 @@ class StrKeyIdDict(UserDict):
     def __delitem__(self, key):
         del self.data[str(key)]
 
+    def __eq__(self, other: object) -> bool:
+        if isinstance(other, StrKeyIdDict):
+            return self.data == other.data
+        if isinstance(other, Mapping) and len(other) == len(self):
+            for key, val in self.items():
+                if key not in other or other[key] != val:
+                    return False
+            return True
+        return False
+
+    def __hash__(self):
+        """not hashable"""
+        return None
+
     @classmethod
     def fromkeys(cls, iterable, value=None):
         d = cls()
